; File:		uart.asm
; Device:	AVR
; Version:	2023-11-02

; === Routine: Send character via UART ===
; Parameter: temp0.
.if ATMEGAXXA_PA
uart_putChar:
	push	temp0
uart_putChar_wait:
	xin		temp0, UCSR0A
	sbrs	temp0, UDRE0
	rjmp	uart_putChar_wait
	pop		temp0
	xout	UDR0, temp0
	ret
.else
.error "Device not supported by library file uart.asm."
.endif

; === Routine: Send string via UART ===
; Parameter: Z pointer, start address of the null-terminated string.
; String length is limited to 255 characters.
uart_putSz:
	push	temp0
	push	temp1
	clr		temp1
uart_putSz_loop:
	lpm		temp0, Z+
	tst		temp0
	breq	uart_putSz_end
	rcall	uart_putChar
	inc		temp1
	cpi		temp1, $FF
	breq	uart_putSz_end
	rjmp	uart_putSz_loop
uart_putSz_end:
	pop		temp1
	pop		temp0
	ret

; === Routine: Send byte in hexadecimal via UART ===
; Send 8-bit unsigned number as two hex digits.
; Parameter: temp0.
uart_putHexByte:
	push	temp0
	swap	temp0
	andi	temp0, 0x0F
	rcall	uart_putHexDigit
	pop		temp0
	push	temp0
	andi	temp0, 0x0F
	rcall	uart_putHexDigit
	pop		temp0
	ret
uart_putHexDigit:
	cpi		temp0, 10
	brlt	uart_putHexDigit_1
	subi	temp0, -('A'-'9'-1)
uart_putHexDigit_1:
	subi	temp0, -'0'
	rcall	uart_putChar
	ret

; === Routine: Send byte in decimal via UART ===
; Suppressing leading zeros.
; Parameter: temp0.
uart_putDecByte:
	push	temp0
	push	temp1
	mov		temp1, temp0
	ldi		temp0, '0'-1
	clt
uart_putDecByte_2:
	inc		temp0
	subi	temp1, 100
	brcc	uart_putDecByte_2
	subi	temp1, low(-100)
	cpi		temp0, '0'
	breq	uart_putDecByte_2z
	set
	rcall	uart_putChar
uart_putDecByte_2z:
	ldi		temp0, '0'-1
uart_putDecByte_1:
	inc		temp0
	subi	temp1, 10
	brcc	uart_putDecByte_1
	subi	temp1, low(-10)
	brts	uart_putDecByte_1ts
	cpi		temp0, '0'
	breq	uart_putDecByte_1z
	set
uart_putDecByte_1ts:
	rcall	uart_putChar
uart_putDecByte_1z:
	subi	temp1, -'0'
	mov		temp0, temp1
	rcall	uart_putChar
	pop		temp1
	pop		temp0
	ret

; === Routine: Send 4-byte number in decimal via UART ===
; Parameters: temp5:temp4:temp3:temp2 contain the number to be sent.
uart_putInteger:
	push	temp0
	push	temp2
	push	temp3
	push	temp4
	push	temp5
; Write the number to UART in decimal ASCII:
	ldi		temp0, '0'-1
	clt							; Used for leading-zero suppression.
uart_putInteger_8:
	inc		temp0
	subi	temp2, low(100_000_000)
	sbci	temp3, byte2(100_000_000)
	sbci	temp4, byte3(100_000_000)
	sbci	temp5, byte4(100_000_000)
	brcc	uart_putInteger_8
	subi	temp2, low(-100_000_000)
	sbci	temp3, byte2(-100_000_000)
	sbci	temp4, byte3(-100_000_000)
	sbci	temp5, byte4(-100_000_000)
	cpi		temp0, '0'
	breq	uart_putInteger_8z
	set
	rcall	uart_putChar
uart_putInteger_8z:
; Value is now <= 99_999_999 = $5_F5_E0_FF.
	ldi		temp0, '0'-1
uart_putInteger_7:
	inc		temp0
	subi	temp2, low(10_000_000)
	sbci	temp3, byte2(10_000_000)
	sbci	temp4, byte3(10_000_000)
	sbci	temp5, byte4(10_000_000)
	brcc	uart_putInteger_7
	subi	temp2, low(-10_000_000)
	sbci	temp3, byte2(-10_000_000)
	sbci	temp4, byte3(-10_000_000)
	sbci	temp5, byte4(-10_000_000)
	brts	uart_putInteger_7ts
	cpi		temp0, '0'
	breq	uart_putInteger_7z
	set
uart_putInteger_7ts:
	rcall	uart_putChar
uart_putInteger_7z:
; Value is now <= 9_999_999 = $98_96_7F, so temp5 contains 0 for sure.
	ldi		temp0, '0'-1
uart_putInteger_6:
	inc		temp0
	subi	temp2, low(1_000_000)
	sbci	temp3, byte2(1_000_000)
	sbci	temp4, byte3(1_000_000)
	brcc	uart_putInteger_6
	subi	temp2, low(-1_000_000)
	sbci	temp3, byte2(-1_000_000)
	sbci	temp4, byte3(-1_000_000)
	brts	uart_putInteger_6ts
	cpi		temp0, '0'
	breq	uart_putInteger_6z
	set
uart_putInteger_6ts:
	rcall	uart_putChar
uart_putInteger_6z:
; Value is now <= 999_999 = $F_42_3F.
	ldi		temp0, '0'-1
uart_putInteger_5:
	inc		temp0
	subi	temp2, low(100_000)
	sbci	temp3, byte2(100_000)
	sbci	temp4, byte3(100_000)
	brcc	uart_putInteger_5
	subi	temp2, low(-100_000)
	sbci	temp3, byte2(-100_000)
	sbci	temp4, byte3(-100_000)
	brts	uart_putInteger_5ts
	cpi		temp0, '0'
	breq	uart_putInteger_5z
	set
uart_putInteger_5ts:
	rcall	uart_putChar
uart_putInteger_5z:
; Value is now <= 99_999 = $1_86_9F.
	ldi		temp0, '0'-1
uart_putInteger_4:
	inc		temp0
	subi	temp2, low(10_000)
	sbci	temp3, byte2(10_000)
	sbci	temp4, byte3(10_000)
	brcc	uart_putInteger_4
	subi	temp2, low(-10_000)
	sbci	temp3, byte2(-10_000)
	sbci	temp4, byte3(-10_000)
	brts	uart_putInteger_4ts
	cpi		temp0, '0'
	breq	uart_putInteger_4z
	set
uart_putInteger_4ts:
	rcall	uart_putChar
uart_putInteger_4z:
; Value is now <= 9_999 = $27_0F, so temp4 contains 0 for sure.
	ldi		temp0, '0'-1
uart_putInteger_3:
	inc		temp0
	subi	temp2, low(1_000)
	sbci	temp3, byte2(1_000)
	brcc	uart_putInteger_3
	subi	temp2, low(-1_000)
	sbci	temp3, byte2(-1_000)
	brts	uart_putInteger_3ts
	cpi		temp0, '0'
	breq	uart_putInteger_3z
	set
uart_putInteger_3ts:
	rcall	uart_putChar
uart_putInteger_3z:
; Value is now <= 999 = $3E7.
	ldi		temp0, '0'-1
uart_putInteger_2:
	inc		temp0
	subi	temp2, low(100)
	sbci	temp3, byte2(100)
	brcc	uart_putInteger_2
	subi	temp2, low(-100)
	sbci	temp3, byte2(-100)
	brts	uart_putInteger_2ts
	cpi		temp0, '0'
	breq	uart_putInteger_2z
	set
uart_putInteger_2ts:
	rcall	uart_putChar
uart_putInteger_2z:
; Value is now <= 99 = $63, so temp3 contains 0 for sure.
	ldi		temp0, '0'-1
uart_putInteger_1:
	inc		temp0
	subi	temp2, 10
	brcc	uart_putInteger_1
	subi	temp2, -10
	brts	uart_putInteger_1ts
	cpi		temp0, '0'
	breq	uart_putInteger_1z
	set
uart_putInteger_1ts:
	rcall	uart_putChar
uart_putInteger_1z:
; temp2 now contains the decimal units.
	subi	temp2, -'0'
	mov		temp0, temp2
	rcall	uart_putChar
; Restore the modified registers:
	pop		temp5
	pop		temp4
	pop		temp3
	pop		temp2
	pop		temp0
	ret

; == Routine uart_getChar ==
; Return: temp0: received byte
uart_getChar:
	xin		temp0, UCSR0A
	sbrs	temp0, RXC0
	rjmp	uart_getChar
; First character received.
	xin		temp0, UDR0
	ret

; == Routine hexChar2num ==
; Parameter:	temp0
; Return:		temp0
hexChar2num:
	sbrc	temp0, 6
	andi	temp0, ~(1<<5)		; small letters -> capitals
	cpi		temp0, '0'
	brlo	hexChar2num_err
	cpi		temp0, '9'+1
	brlo	hexChar2num_09
	cpi		temp0, 'A'
	brlo	hexChar2num_err
	cpi		temp0, 'F'+1
	brlo	hexChar2num_AF
hexChar2num_err:
	ldi		temp0, $FF
	rjmp	hexChar2num_end
hexChar2num_09:
	subi	temp0, '0'
	rjmp	hexChar2num_end
hexChar2num_AF:
	subi	temp0, 'A'-10
hexChar2num_end:
	ret

; == Routine: read byte as two hex chars from UART ==
; Return: temp0
uart_getHexByte:
	push	temp1
	rcall	uart_getChar
	rcall	hexChar2num
	mov		temp1, temp0
	rcall	uart_getChar
	rcall	hexChar2num
	swap	temp1
	or		temp0, temp1
	pop		temp1
	ret
	